/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Should you need to contact me, the author, you can do so by
 * e-mail - mail your message to Viktor T. Toth at <vttoth@vttoth.com>.
 *
 * ACKNOWLEDGEMENTS
 *
 * Many thanks to Paul Godfrey whose work provided the basis for this
 * implementation.
 *
 * Updated 2007 by Robert Ham for tkresolver
 *
 * Copyright (C) 2007 Robert Ham <rah@bash.sh> for tkresolver
 * Copyright (C) Viktor T. Toth at <vttoth@vttoth.com>
 *
 */

#include "lanczos.h"
#include "matrix.h"
#include "complex.h"
#include "real.h"
#include "rational.h"

using namespace TkResolver;
using namespace MP;


template<typename T>
T
combination(const Integer::Number &n, const Integer::Number &k)
{
  if (k < 0)
    return 0;
  
  T r(1);
  for (Integer::Number l(1); l <= n-k; ++l)
    {
      r *= n + 1 - l;
      r /= l;
    }
  
  return r;
};

template <typename T>
MP::Matrix<T>
make_B(const Integer::Number &n)
{
  MP::Matrix<T> B(n, n);
  int r, c;
  
  for (c = 0; c < n; ++c)
    B[0][c] = 1;
    
  for (r = 1; r < n; ++r)
    for (c = 0; c < n; c++)
      {
	B[r][c] = combination<T>(r+c-1, c-r);
	
	if ((r+c)&1)
	  B[r][c] = -B[r][c];
      }
  
  return B;
}


template <typename T>
MP::Matrix<T>
make_C(const Integer::Number &n)
{
  MP::Matrix<T> C(n, n);
  
  for (int i(1); i < n; ++i)
    for (int j(0); j <= i; ++j)
      {
	C[i][j] = 0;
	
	for (int k(0); k <= i; ++k)
	  C[i][j] += combination<T>(2*i, 2*k) * combination<T>(k, k+j-i);
	
	if ((i-j)&1)
	  C[i][j] = -C[i][j];
      }
  
  C[0][0] = 1/2;
  
  return C;
}

template <typename T>
MP::Matrix<T>
make_D(const Integer::Number &n)
{
  MP::Matrix<T> D(n, n);

  D[0][0] = 1;
  D[1][1] = -1;
  
  for (Integer::Number i(2); i < n; ++i)
    {
      D[i][i] = D[i-1][i-1] * (2 * (2*i-1));
      D[i][i] /= (i - 1);
    }
  
  return D;
}

static Real::Matrix
make_F(const Real::Number &g, const Integer::Number &n)
{
  Real::Matrix F(n, 1);
  
  for (Integer::Number a(0); a < n; ++a)
    {
      F[a][0] = 2;
      for (Integer::Number i(a + 1); i <= 2 * a; ++i)
	{
	  F[a][0] *= i;
	  F[a][0] /= 4;
	}
      F[a][0] *= Real::exp(static_cast<Real::Number>(a) + g + .5);
      F[a][0] /= Real::pow(static_cast<Real::Number>(a) + g + .5, a.get_ui());
      F[a][0] /=      sqrt(static_cast<Real::Number>(a) + g + .5);
    }
  
  return F;
}


Lanczos::Context:: Context(const Integer::Number &n, const Real::Number &g, unsigned long precision)
  : _n(n),
    _g(g),
    _prec(precision),
    _P(n, 1)
{
  unsigned long old_prec(mpf_get_default_prec());
  mpf_set_default_prec(precision);
  
  /*  Rational::Matrix B( make_B<Rational::Number>(n) );
  Rational::Matrix C( make_C<Rational::Number>(n) );
  Rational::Matrix D( make_D<Rational::Number>(n) );
  Real::Matrix F( make_F(g, n) );
  
  _P = static_cast<Real::Matrix>(D)
    *static_cast<Real::Matrix>(B)
    *static_cast<Real::Matrix>(C)
    *F;*/
  
  Rational::Matrix B( make_B<Rational::Number>(n) );
  Rational::Matrix C( make_C<Rational::Number>(n) );
  Rational::Matrix D( make_D<Rational::Number>(n) );
  Real::Matrix F( make_F(g, n) );
  
  /*  _P = static_cast<Real::Matrix>(D)
    *static_cast<Real::Matrix>(B)
    *static_cast<Real::Matrix>(C)
    *F;*/
  _P = static_cast<Real::Matrix>(D*B*C)*F;
  
  mpf_set_default_prec(old_prec);
}

const Integer::Number &
Lanczos::Context:: get_n() const
{
  return _n;
}

const Real::Number &
Lanczos::Context:: get_g() const
{
  return _g;
}

const Real::Matrix &
Lanczos::Context:: get_P() const
{
  return _P;
}

const unsigned long &
Lanczos::Context:: get_prec() const
{
  return _prec;
}


Complex::Number
Lanczos:: lngamma(const Complex::Number &z_, const Lanczos::Context &ctx)
{
  unsigned long old_prec(mpf_get_default_prec());
  mpf_set_default_prec(ctx.get_prec());
  
  Real::Number zRe(z_.real());
  Real::Number zIm(z_.imag());
  
  bool s = false;
  if (zRe < 0)
    {
      zRe = 1 - zRe;
      s = true;
    }
  Complex::Number z(zRe - 1, zIm), r;
  
  Complex::Matrix Z(1, ctx.get_n());
  Complex::Matrix R(1, 1);
  
  Z[0][0] = Real::Number(1);
  for (Integer::Number i(1); i < ctx.get_n(); ++i)
    Z[0][i] = 1 / (z + i);
  
  R = Z * ctx.get_P();
  r = z + ctx.get_g() + .5;
  Complex::Number r1( Complex::log(R[0][0]) );
  Complex::Number r2( (z+.5) * Complex::log(r));
  Complex::Number r3(r1);
  //  r = Complex::log(R[0][0]) + (z+.5) * Complex::log(r) - r;
  r = r1 + r2 - r3;
  if (s)
    r = log(Real::pi) - r - Complex::log(-Complex::sin(Real::pi * z));
  
  mpf_set_default_prec(old_prec);
  return r;
}

Complex::Number
Lanczos:: gamma(const Complex::Number &z, const Lanczos::Context &ctx)
{
  return Complex::exp(lngamma(z, ctx));
}

Real::Number
Lanczos:: lngamma(const Real::Number &z_, const Lanczos::Context &ctx)
{
  unsigned long old_prec(mpf_get_default_prec());
  mpf_set_default_prec(ctx.get_prec());
  
  Real::Number z(z_);
  bool s = false;
  if (z < 0.0)
    {
      z = 1.0 - z;
      s = true;
    }
  
  Real::Number r;
  Real::Matrix Z(1, ctx.get_n());
  Real::Matrix R(1, 1);
  
  Z[0][0] = Real::Number(1.0);
  for (Integer::Number i(1); i < ctx.get_n(); ++i)
    Z[0][i] = 1.0 / (z + static_cast<Real::Number>(i));
  
  R = Z * ctx.get_P();
  r = z + ctx.get_g() + .5;
  std::cerr << "finding log of: " << R[0][0] << std::endl;
  Real::Number r1( Real::log(R[0][0]) );
  std::cerr << "found log of: " << R[0][0] << std::endl;
  Real::Number r2( (z+.5) * Real::log(r));
  Real::Number r3(r1);
  //  r = Complex::log(R[0][0]) + (z+.5) * Complex::log(r) - r;
  r = r1 + r2 - r3;
  if (s)
    r = log(Real::pi) - r - Real::log(-Real::sin(Real::pi * z));
  
  mpf_set_default_prec(old_prec);
  return r;
}

Real::Number
Lanczos:: gamma(const Real::Number &f, const Lanczos::Context &ctx)
{
  return Real::exp(lngamma(f, ctx));
}

