/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#include <unistd.h>

#include <iostream>
#include <cmath>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/tkintl.h"
#include "libtkcommon/maths.h"
#include "libtkcommon/real_time.h"
#include "randomness.h"
#include "retriever_registry.h"
#include "random_types.h"

using namespace TkResolver;
using namespace Random;

#define MAX_BUFFER_SIZE 16384

Randomness:: Randomness(const String &retriever_type, bool deterministic, unsigned buffer_size, bool record_use)
  : _run_state(Stopped),
    _retriever_type(retriever_type),
    _retrieving(0),
    _retrieved(0),
    _buffer_size(buffer_size > MAX_BUFFER_SIZE ? MAX_BUFFER_SIZE : buffer_size),
    _bit_index(0),
    _used_bits(record_use ? new Bits : 0)
{
  if(!Glib::thread_supported())
    Glib::thread_init();

  unsigned preferred_block_size = RetrieverRegistry::get_preferred_block_size(_retriever_type);
  
  if (deterministic && buffer_size < MAX_BUFFER_SIZE)
    _threshold = 0;
  else if (preferred_block_size > _buffer_size)
    _threshold = _buffer_size / 2;
  else
    _threshold = _buffer_size - preferred_block_size;
    
  /*  WDEBUG_OUT << L"Randomness"
	     << L"; outcomes: " << _outcomes
	     << L"; count: " << count
	     << L"; tree depth: " << _decision_tree.get_depth()
	     << L"; tree outcomes: " << _decision_tree.tree_outcomes()
	     << L"; expected tries: " << _decision_tree.expected_tries()
	     << L"; buffer size: " << _buffer_size
	     << L"; threshold: " << _threshold
	     << std::endl;*/

  pthread_mutex_init (&_join_lock, NULL);
  pthread_cond_init (&_join_signal, NULL);

  // fill buffer
  Retriever * retriever = RetrieverRegistry::new_retriever(_retriever_type, _buffer_size);
  retriever->start();
  _retrieving = _buffer_size;
  Buffer * buffer = retriever->join();
  delete retriever;
  append_buffer(buffer);
  delete buffer;

  // fill first randomness
  _randomness = _buffer.back();
  _buffer.pop_back();
  
  // kick off join thread
  _run_state = Running;
  
  //  _thread = Glib::Thread::create(sigc::mem_fun(*this, &Retriever::run), true);
  pthread_create(&_join_thread, NULL, low_join_run, this);
}

Randomness:: ~Randomness()
{
  _run_state = Stopped;
  pthread_cond_signal(&_join_signal);

  int err = pthread_join(_join_thread, NULL);
  if (err)
    { 
      std::wcerr  << __FUNCTION__
		  << L": could not join join thread: ";
      std::cerr   << strerror (err)
		  << std::endl;
      throw UnixError (_("could not join to join thread"), err);
    }

  /*  WDEBUG_OUT << "decision retries: "
	     << _decision_tree.get_retries()
	     << " ("
	     << (_count * (_decision_tree.expected_tries() - 1.0))
	     << " expected)"
	     << std::endl;*/
}

void
Randomness:: lock()
{
  //  WDEBUG_OUT << "locking" << std::endl;
  pthread_mutex_lock(&_join_lock);
  //  WDEBUG_OUT << "locked" << std::endl;
}

void
Randomness:: unlock()
{
  pthread_mutex_unlock(&_join_lock);
  //  WDEBUG_OUT << "unlock" << std::endl;
}

void *
Randomness:: low_join_run(void * data)
{
  Randomness * randomness = static_cast<Randomness *>(data);
  return randomness->join_run();
}

void
Randomness:: append_buffer(const Buffer * buffer)
{
  lock();
  _buffer.insert<Buffer::const_iterator>(_buffer.begin(), buffer->begin(), buffer->end());
  _retrieving -= buffer->size();
  _retrieved += buffer->size();
  unlock();
}

void *
Randomness:: join_run()
{
  Retriever * retriever;
  Buffer * buffer;

  thread_set_real_time(30);

  while (_run_state != Stopped)
    {
      lock();
      if (_retrievers.empty())
	{
	  pthread_cond_wait(&_join_signal, &_join_lock);
	}

      if (_retrievers.empty())
	{
	  unlock();
	  break;
	}

      retriever = _retrievers.front();
      _retrievers.pop_front();
      

      unlock();

      buffer = retriever->join();
      delete retriever;
      
      append_buffer(buffer);
      delete buffer;
    }

  return NULL;
}

void
Randomness:: retrieve(unsigned count)
{
  Retriever * retriever = RetrieverRegistry::new_retriever(_retriever_type, count);
  retriever->start();

  lock();
  _retrieving += count;
  _retrievers.push_back(retriever);
  unlock();

  pthread_cond_signal(&_join_signal);
}

unsigned
Randomness:: retrieved()
{
  unsigned retrieved;
  lock();
  retrieved = _retrieved;
  unlock();
  return retrieved;
}

void
Randomness:: check_threshold()
{
  if (_threshold == 0)
    return;

  unsigned buffered;
  unsigned retrieving;
  
  lock();
  retrieving = _retrieving;
  buffered = _buffer.size();
  unlock();

  /* WDEBUG_OUT << L"buffered: " << buffered
  	     << L"; currently retrieving: " << retrieving
  	     << L"; threshold: " << _threshold
  	     << std::endl;*/
  if (buffered + retrieving <= _threshold)
    {
      WDEBUG_OUT << L"getting more randomness (" << _buffer_size - retrieving - buffered << L")" << std::endl;
      retrieve(_buffer_size - retrieving - buffered);
      usleep(100);
    }
}

bool
Randomness:: get_bit()
{
  bool bit;

  bit = _randomness & (1 << _bit_index);


  if (_used_bits)
    _used_bits->push_back (bit);

  ++_bit_index;
  
  if (_bit_index >= BITS_PER_CHAR )
    {
      _bit_index = 0;
      
      if (_threshold != 0 || retrieved() < _buffer_size)
	{
	  lock();
      
	  if (_buffer.empty())
	    {
	      unlock();
	      throw std::underflow_error("out of randomness");
	    }
      
	  _randomness = _buffer.back();
	  _buffer.pop_back();
      
	  unlock();

	  check_threshold();
	}
    }
  
  return bit;
}

const Bits&
Randomness:: get_used_bits() const
{
  const static Bits null_bits;
  
  if (!_used_bits)
    return null_bits;
  
  return *_used_bits;
}



unsigned long
Randomness:: get_number(unsigned long upper, const String &retriever_type)
{
  const unsigned short digits = std::numeric_limits<unsigned long>::digits;
  unsigned long bits = 0;

  Randomness * randomness = new Randomness(retriever_type, true, digits / BITS_PER_CHAR, false);
  

  for (unsigned short i = 0; i < digits; ++i)
    {
      if (randomness->get_bit())
	bits |= (1 << i);
    }
  /*
  for (unsigned short i = 0; i < digits; ++i)
    {
      bits <<= 1;
      bits |= randomness->get_choice();
      }
  */

  delete randomness;

  const unsigned long max( std::numeric_limits<unsigned long>::max() );
  unsigned long value(
    static_cast<unsigned long>(
      static_cast<long double>(upper) * (bits / (max + 1.0)) ) );

  return value;
}


/* EOF */
