/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * GPL all the way, baby :)
 *
 */

#define __STDC_LIMIT_MACROS

#include <stdlib.h>
#include <stdint.h>

#include <iostream>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "display.h"
#include "tkresolver.h"
#include "sdl.h"

using namespace TkResolver;


unsigned
Display:: calc_total_lines(unsigned total_time, double speed, unsigned height)
{
  return static_cast<unsigned>( total_time * speed * height );
}


Display:: Display (unsigned height,
		   unsigned width,
		   unsigned trace_count,
		   unsigned total_time,
		   bool     scrolled,
		   double   speed,
		   double   horizon)
  : _trace_count(trace_count),
    _scrolling (scrolled),
    _last_y (-1),
    _print_count(0),
    _x_history(0)
{
  SDL_ShowCursor(SDL_DISABLE);
  
  _sdl_display = SDL_SetVideoMode(width, height, 8, SDL_ANYFORMAT|SDL_FULLSCREEN|SDL_HWSURFACE);
  if (!_sdl_display) 
    throw std::runtime_error(std::string("could not initialise display: ") + SDL_GetError());
  
  // wait for the monitor to switch
  sleep (2);
  
  
  if (_scrolling)
    {
      _total_lines = calc_total_lines(total_time, speed, get_height());
      DEBUG_OUT << "total lines for scrolling: " << _total_lines << std::endl;
      _horizon     = static_cast<unsigned>( horizon * get_height() );
    }

  if (!_scrolling || _total_lines < get_height())
    {
      _total_lines = _horizon = get_height();
      DEBUG_OUT << "total lines for non-scrolling: " << _total_lines << std::endl;
      _scrolling = false;
    }

  if (_scrolling)
    _x_history = new LineList();
  
  _white = SDL_MapRGB (_sdl_display->format, UINT8_MAX, UINT8_MAX, UINT8_MAX);
  _black = SDL_MapRGB (_sdl_display->format, 0, 0, 0);
  _red   = SDL_MapRGB (_sdl_display->format, UINT8_MAX, 0, 0);

};

Display::~Display()
{
  /*SDL_FreeSurface(_sdl_display);*/

  /* Seems to be the only way to do this */
  SDL::reset();
}

unsigned
Display:: get_height() const
{
  return _sdl_display->h;
};

unsigned
Display:: get_width() const
{
  return _sdl_display->w;
};

unsigned
Display:: get_total_lines() const
{
  return _total_lines;
}

/*
const Line &
Display:: get_start_line() const
{
  return _start_x;
}
*/

void
Display:: lock()
{
  if (SDL_MUSTLOCK(_sdl_display))
    {
      int err = SDL_LockSurface(_sdl_display);
      if (err)
	throw std::runtime_error(std::string("could not lock SDL surface: ") + SDL_GetError());
    }
}

void
Display:: unlock()
{
  if (SDL_MUSTLOCK(_sdl_display))
    SDL_UnlockSurface(_sdl_display);
}

int
Display:: truncate_x(int x)
{
  if (x < 0)
    x = 0;
  else if (x > get_width())
    x = get_width();

  return x;
}

Line
Display:: calc_start_line() const
{
  Line start_line;
  unsigned trace_separation( get_width() / (_trace_count + 1) );
  for (OffsetIndex i = 0; i < _trace_count; ++i)
    start_line.push_back( trace_separation * (i + 1) );
  
  return start_line;
}

int
Display:: get_last_x(OffsetIndex index) const
{
  return _x_history->back()[index];
}

void
Display:: get_x_limits(OffsetIndex index, int &min, int &max)
{
  int x;
  
  if (_x_history == 0)
    throw std::logic_error("cannot get x limits while not scrolling");
  
  if (_x_history->size() == 0)
    throw std::logic_error("cannot get x limits from empty history");

  min = max = _x_history->front()[index];

  for (LineList::iterator iter( ++(_x_history->begin()) );
       iter != _x_history->end();
       ++iter)
    {
      x = (*iter)[index];
      
      if (x > max)
	max = x;
      else if (x < min)
	min = x;
    }
}

void
Display:: print_line (const Line &line)
{
  OffsetIndex i;
  SDL_Rect rect;
  SDL_Rect blits[_trace_count];
  
  rect.w = rect.h = 1;
  
  
  if (!_scrolling ||
      _last_y < _horizon ||
      _print_count >= _total_lines - (get_height() - _horizon))
    {
      rect.y = _last_y + 1;
    }
  else
    {
      rect.y = _last_y;
    }
  
  
  if (_last_y > 0 && rect.y == _last_y)
    {
      int x_max, x_min;

      SDL_Rect src;
      src.y = 1;
      src.h = _horizon;
      
      for (i = 0; i < _trace_count; ++i)
	{
	  // FIXME: check whether histories overlap on x axis
	  get_x_limits(i, x_min, x_max);

	  src.x = x_min;
	  src.w = x_max - x_min + 1;
	  
	  blits[i] = src;
	  blits[i].y = 0;
	  
	  int err = SDL_BlitSurface(_sdl_display, &src, _sdl_display, blits + i);
	  if (err)
	    throw std::runtime_error(std::string("could not blit for scroll: ") + SDL_GetError());
	  
	  
	  if (line[i] != get_last_x(i))
	    {
	      SDL_Rect old( rect );
	      old.x = get_last_x(i);
	      lock();
	      SDL_FillRect(_sdl_display, &old, _black);
	      unlock();
	    }
	}
    }
  
  lock();
  for (i = 0; i < _trace_count; ++i)
    {
      rect.x = line[i];
      SDL_FillRect(_sdl_display, &rect, _white);
    }
  unlock();
  
  if (rect.y == _last_y)
    {
      for (i = 0; i < _trace_count; ++i)
	{
	  SDL_UpdateRect(_sdl_display, blits[i].x, 0, blits[i].w, rect.y);
	  
	  if (line[i] != get_last_x(i))
	    SDL_UpdateRect(_sdl_display, get_last_x(i), rect.y, 1, 1);
	}
    }
  
  for (i = 0; i < _trace_count; ++i)
    SDL_UpdateRect(_sdl_display, line[i], rect.y, 1, 1);
  
  
  if (_scrolling)
    {
      if (rect.y == _last_y && _x_history->size() > 0)
	_x_history->erase(_x_history->begin());
      
      _x_history->push_back(line);
    }

  _last_y = rect.y;
  ++_print_count;
}

Display::BlockList
Display:: create_count_down_blocks(unsigned count)
{
  const double phi = 1.61803398874989484820458683436563811772030917980576;
  
  unsigned total_height = static_cast<unsigned>( get_height() * 5.0/7.0 );
  unsigned top          = (get_height() - total_height) / 2;
  unsigned height       = total_height / (count*2 - 1);
  unsigned width        = static_cast<unsigned>( height * phi );
  unsigned left         = static_cast<unsigned>( get_width()/2.0 - width/2.0 );
  
  BlockList list;
  for (unsigned i = 0; i < count; ++i)
    {
      SDL_Rect block;
      block.x = left;
      block.y = top + height*2*i;
      block.h = height;
      block.w = width;

      list.push_back(block);
    }

  return list;
}

void
Display:: count_down (unsigned count)
{
  BlockList blocks(create_count_down_blocks(count));

  // draw blocks
  lock();
  for (BlockList::iterator iter(blocks.begin());
       iter != blocks.end();
       ++iter)
    {
      SDL_FillRect(_sdl_display, &(*iter), _red);
    }
  unlock();
  
  SDL_UpdateRect(_sdl_display, blocks.front().x, blocks.front().y, blocks.front().w, blocks.back().y + blocks.back().h - blocks.front().y);

  // cover them up again
  for (BlockList::reverse_iterator iter(blocks.rbegin());
       iter != blocks.rend();
       ++iter)
    {
      SDL_Delay(1000);

      lock();
      SDL_FillRect(_sdl_display, &(*iter), _black);
      unlock();
      
      SDL_UpdateRect(_sdl_display, iter->x, iter->y, iter->w, iter->h);
    }
  
  SDL_Delay(1000);
}
